CUSTOMER_UNITS_FILE_NAME = "customer_units.json"
PROCESS_LINES_FILE_NAME = "process_lines.json"
PRODUCTION_UNITS_FILE_NAME = "production_units.json"
PRODUCTS_FILE_NAME = "products.json"
SALE_UNITS_FILE_NAME = "sale_units.json"
SCHEDULING_UNITS_FILE_NAME = "scheduling_units.json"
STORAGE_UNITS_FILE_NAME = "storage_units.json"
OBS_DEFINE_FILE_NAME = "obs_define.json"
VARIABLE_MAP_FILE_NAME = "variable_map.json"
VARIABLE_DEFINE_FILE_NAME = "variable_define.json"

CUSTOMER_UNITS_ID_PREFIX = "CustomerUnit"
PROCESS_LINES_ID_PREFIX = "ProcessLine"
PRODUCTION_UNITS_ID_PREFIX = "ProductionUnit"
SUBSTANCE_ID_PREFIX = "Substance"
SALE_UNITS_ID_PREFIX = "SaleUnit"
SCHEDULING_UNITS_ID_PREFIX = "SchedulingUnit"
STORAGE_UNITS_ID_PREFIX = "StorageUnit"
PARALLEL_COLLECTION_UNITS_ID_PREFIX = "ParallelCollectionUnit"














